use std::error::Error;

use argh::FromArgs;
use notify_rust::{Hint, Notification};
use pulsectl::controllers::{DeviceControl, SinkController};

/// Control the volume of the default sink
#[derive(Clone, Copy, Debug, FromArgs)]
struct Args {
    #[argh(subcommand)]
    command: Subcommand,
}

#[derive(Debug, Clone, Copy, PartialEq, FromArgs)]
#[argh(subcommand)]
enum Subcommand {
    Up(UpArgs),
    Down(DownArgs),
    Mute(MuteArgs),
    Get(GetArgs),
    Set(SetArgs),
}

/// Increase the volume
#[derive(Debug, Clone, Copy, PartialEq, FromArgs)]
#[argh(subcommand, name = "up")]
struct UpArgs {}

/// Decrease the volume
#[derive(Debug, Clone, Copy, PartialEq, FromArgs)]
#[argh(subcommand, name = "down")]
struct DownArgs {}

/// Mute the sink
#[derive(Debug, Clone, Copy, PartialEq, FromArgs)]
#[argh(subcommand, name = "mute")]
struct MuteArgs {}

/// Get the current volume
#[derive(Debug, Clone, Copy, PartialEq, FromArgs)]
#[argh(subcommand, name = "get")]
struct GetArgs {
    /// output bare value
    #[argh(switch, short='b')]
    bare: bool,
}

/// Set the volume
#[derive(Debug, Clone, Copy, PartialEq, FromArgs)]
#[argh(subcommand, name = "set")]
struct SetArgs {
    #[argh(positional)]
    volume: i32,
}

enum Volume {
    Muted,
    Percent(i32),
}

fn main() -> Result<(), Box<dyn Error>> {
    let args: Args = argh::from_env();

    let mut handler = SinkController::create()?;

    let mut default_device = handler.get_default_device()?;
    let muted = default_device.mute;
    let volume_step = pulse::volume::Volume(pulse::volume::Volume::NORMAL.0 / 20);

    match args.command {
        Subcommand::Up(_) => {
            if default_device.volume.avg().0 < pulse::volume::Volume::NORMAL.0 {
                default_device.volume.inc_clamp(volume_step, pulse::volume::Volume::NORMAL);
                handler.set_device_volume_by_index(default_device.index, &default_device.volume);
            }
        }
        Subcommand::Down(_) => {
            default_device.volume.decrease(volume_step);
            handler.set_device_volume_by_index(default_device.index, &default_device.volume);
        }
        Subcommand::Mute(_) => {
            handler.set_device_mute_by_index(default_device.index, !muted);
        }
        Subcommand::Get(GetArgs { bare }) => {
            if bare {
                print!("{}", ((default_device.volume.avg().0 as f64 / pulse::volume::Volume::NORMAL.0 as f64) * 100.0).round());
            } else {
                println!("Volume: {}", default_device.volume.avg());
            }
            return Ok(());
        }
        Subcommand::Set(SetArgs { volume }) => {
            let mut current_volume = default_device.volume;
            let new_volume = pulse::volume::Volume(volume_from_percent(volume) as u32);
            let volumes = current_volume.set(2, new_volume);
            handler.set_device_volume_by_index(default_device.index, volumes);
        }
    }
    let v = if handler.get_default_device()?.mute {
        Volume::Muted
    } else {
        Volume::Percent(volume_to_percent(
            handler.get_default_device()?.volume.avg().0 as i32,
        ))
    };
    notify_volume(v);

    Ok(())
}

fn icon_name(volume: Volume) -> String {
    let name = match volume {
        Volume::Muted => "muted",
        Volume::Percent(x) if x <= 33 => "low",
        Volume::Percent(x) if x <= 67 => "medium",
        _ => "high",
    };

    format!("audio-volume-{}-symbolic", name)
}

fn notify_volume(new_volume: Volume) {
    let value = match new_volume {
        Volume::Muted => 0,
        Volume::Percent(new_volume) => new_volume,
    };
    Notification::new()
        .summary("Volume")
        .icon(icon_name(new_volume).as_str())
        .appname("pamedia")
        .hint(Hint::SoundName("audio-volume-change".into()))
        .hint(Hint::Custom("synchronous".into(), "audio".into()))
        .hint(Hint::CustomInt("value".into(), value))
        .show()
        .expect("error showing notification");
}

fn volume_from_percent(volume: i32) -> f64 {
    (volume as f64) * (f64::from(pulse::volume::Volume::NORMAL.0) / 100.0)
}

fn volume_to_percent(volume: i32) -> i32 {
    ((volume as f64) / (f64::from(pulse::volume::Volume::NORMAL.0)) * 100.0).round() as i32
}
